;
        TITLE 'PIC12F675 ADC Demo'
;
; Copyright Dennis Clark and TTT Enterprises 10/2009.  This code is free to 
; use and distribute according to the Free Software Copyleft rules and 
; conventions.  You may not publish or sell it without written permission 
; from me, the author.
;

#include <P12F675.INC>

; Here we will use the chip reset (~MCLR), nor the watchdog, and we'll use
; the chip's internal 4MHz RC oscillator since we don't need any kind of real
; accuracy in this program.

	__CONFIG        ( _MCLRE_ON & _BODEN_OFF & _CP_OFF & _WDT_OFF & _INTRC_OSC_NOCLKOUT )

;==========================================================================
; MACROS
;
;  These are convenience macros that simplify the code.
;==========================================================================

BANK0	macro						; switch to BANK 0 RAM
	bcf	STATUS,RP0
	endm
	
BANK1	macro						; switch to BANK 1 RAM
	bsf	STATUS,RP0
	endm
		
ADC_GO	macro						; start an ADC conversion
	bsf	ADCON0,GO_DONE
	endm
	
;==========================================================================
; Program Variables and register defines.
;==========================================================================

; Port definitions and other constants
#define SCONFIG		B'11011000'   	; inputs= 3 & 4, outputs 0, 1, 2, 5
#define OPTS		B'10000000'		; no pull ups, no Timer prescaler
#define ANALOG		0x18			; FOSC/8, use AN3 (GP4) for ADC input
#define ANCONF		0x0D			; Left justify, Vcc reference, channel 3
#define CMPOFF		0x07			; Turn the analog comparitor off!

	LIST							; Start the assembly listing here.
	
; Register locations for important stuff, easier to remember than reg. #'s
pattern     EQU     0x20        	; LED state pattern

;==========================================================================
; Setup Code     
;
;  Configure the Outputs, ADC and PIC hardware. The <== lines MUST be done
; when we are using the internal RC oscillator, this sets up the factory
; calibration of the oscillator.  When we set up the ADC module we need
; to make sure that the system has the proper time to correctly read an
; analog value.  In this case we are dividing the 4MHz clock by 8 to get
; a Tad of 1.6us or greater.  In this case 1/(4MHz/8) = 2us.  Carefully
; read Section 7 of the 12F675 data sheet to understand this.
;==========================================================================

start
    ORG     0x000					; Absolute address
    BANK1
	call	0x3FF					; <== Get the cal value
	movwf   OSCCAL          		; <== Store the factory osc. calibration value
	
	movlw   SCONFIG					; Set how we want to configure I/O pins
	movwf	TRISIO					; Configure input and output pins
	movlw   OPTS		     		; Set OPTION bits as we have defined them
	movwf	OPTION_REG				; Implement OPTION register bits
	movlw	ANALOG
	movwf	ANSEL					; set ADC clock and inputs
	BANK0
	movlw	ANCONF
	movwf	ADCON0					; configure ADC and start converter
	clrf	GPIO					; clear GPIO
	movlw	CMPOFF
	movwf	CMCON					; turn comparitor off
	
;==========================================================================
; Main code loop 
;
;  First take and ADC reading, we're only interested in the upper 8 bits,
; we set the "left justify" bit in ADCON0 so that ADRESH (upper 8 bits of 
; the 10 bit result) gets the analog value divided by 4.  We have to wait 
; for the GO_DONE bit in ADCON0 to clear to show that we have a new value.
; Take this reading, divide that by 16 to get a 4 bit value and then adjust
; the result to match the GPIO output bits pattern of GP0, 1, 2 and 5 
; digital outputs.  We're using GP4 as our analog input and GP3 can only 
; be an input, that port can't be an output, so we'll use it as an enable
; line.  Optionally, you could use it as a reset pin, but that won't do 
; much for this project.
;
; Cute option, you could also divide by 16 by using 'swapf pattern,f' which
; swaps the upper and lower nibble of a byte.  This only works though when
; you want to divide by 16.
;
;  The end result is that if you arrage the LED's in their proper binary 
; order, you'll see the count from 0 to 15 (0000 to 1111) reflect the 
; analog input.  In our case, we'll put a pot (variable resistor) on that 
; line so that the value is from 0 to 5V.  
;==========================================================================

loop
	ADC_GO							; fire off an ADC sample
	
wait								; stay in this loop until ADCON0, GO_DONE bit clears
	btfsc	ADCON0,GO_DONE
	goto	wait					; not yet
	
	movf	ADRESH,W				; Get the upper 8 bits of the conversion (divide by 4)
	movwf	pattern					; and save it to the output pattern register
	bcf		STATUS,C
	movlw	0xF0					; clear lower nibble
	andwf	pattern,f				; so that the rrf won't put '1' into the carry bit
	rrf		pattern,f				; These 
	rrf		pattern,f				;  four
	rrf		pattern,f				;   shifts
	rrf		pattern,f				;    divide this number by 16
	btfss	pattern,3				; check to see if bit 3 is set
	goto	no_change				; it wasn't
	
change
	bsf		pattern,5				; set bit for GP5 ping
	bcf		pattern,3				; and clear this one
	
no_change
	movf	pattern,w				; get the results of our math			
	movwf	GPIO					; and output this to the LED's out there
	goto	loop					; and do it again!
	
		END            				; directive 'end of program'

