;
        TITLE 'PIC12F675 ADC Controls Servo Demo'
;
; Copyright Dennis Clark and TTT Enterprises 10/2009.  This code is free to 
; use and distribute according to the Free Software Copyleft rules and 
; conventions.  You may not publish or sell it without written permission 
; from me, the author.
;

#include <P12F675.INC>

; Here we will use the chip reset (~MCLR), but not the watchdog, and we'll use
; the chip's internal 4MHz RC oscillator since we don't need any kind of real
; accuracy in this program.

	__CONFIG        ( _MCLRE_ON & _CP_OFF & _WDT_OFF & _INTRC_OSC_NOCLKOUT )

;==========================================================================
; MACROS
;
;  These are convenience macros that simplify the code.  The servo bit set
; macros are there because of a potential for a "read modify write" issue 
; with low and mid-range PIC devices when flipping bits in a GPIO register.
; To avoid any chance of that we use a "shadow" register so we write all
; of the bits out at the same time.
;==========================================================================

BANK0	macro						; switch to BANK 0 RAM
	bcf		STATUS,RP0
	endm
	
BANK1	macro						; switch to BANK 1 RAM
	bsf		STATUS,RP0
	endm
		
ADC_GO	macro						; start an ADC conversion
	bsf		ADCON0,GO_DONE
	endm
	
SERVO_SET macro						; set the servo bit
	bsf		pattern,5
	movf 	pattern,W
	movwf	GPIO					; output the bit pattern
	endm
	
SERVO_CLR	macro					; clear the servo bit
	bcf		pattern,5
	movf 	pattern,W
	movwf	GPIO					; output the bit pattern
	endm
;==========================================================================
; Program Variables and register defines.
;==========================================================================

; Port definitions and other constants
#define SCONFIG		B'11011000'   	; inputs= 3 & 4, outputs 0, 1, 2, 5
#define OPTS		B'10000010'		; no pull ups, /8 prescaler on TMR0
#define ANALOG		0x18			; FOSC/8, use AN3 (GP4) for ADC input
#define ANCONF		0x0D			; Left justify, Vcc reference, channel 3
#define CMPOFF		0x07			; Turn the analog comparitor off!
#define T1SETTING	0x01			; TMR1 just turn it on
#define T1MSB		0xB5			; TMR1 19ms upper byte setting
#define T1LSB		0xC7			; and the lower byte setting

	LIST							; Start the assembly listing here.
	
; Register locations for important stuff, easier to remember than reg. #'s
light	    EQU     0x20        	; LED state pattern
STATUS_TMP	EQU		0x21			; IRQ save
W_TMP		EQU		0x22			; as above
servo		EQU		0x23			; new servo value
pattern		EQU		0x24			; servo output shadow register
V255		EQU		0x25			; holds value 255
cur_servo	EQU		0x26			; the current servo value used

start
    ORG     0x000					; Absolute address
    goto	init					; jump around ISR
    
    ORG		0x004
isr									; here is the common IRQ vector
	;======================================================================
	; This ISR (interrupt service routine) looks scary, but it isn't
	; really.  The PIC is REALLY easy to work with when it comes to
	; interrupts, especially in assembly.  The pre and post amble save
	; registers (W and STATUS) that needs to be saved.  
	; What happens here:
	; First we'll get an interrupt that is the duration of the servo pulse.
	; When that happens we'll disable the TMR0 interrupt and then set up 
	; and enable the TMR1 interrupt.  When the TMR1 interrupt happens, we
	; disable the TMR1 interrupt and then set up and enable the TMR0 
	; interrupt.  Note everything that goes on here, none of it is 
	; optional!
	;======================================================================
	movwf 	W_TMP 					; copy W to temp register,
	swapf 	STATUS,W 				; swap status to be saved into W
	BANK0							; current bank is stored in STATUS
	movwf 	STATUS_TMP 				; save status to bank 0 register

	btfss	PIR1,TMR1IF				; TMR1 int?
	goto	do_TMR0					; no, must be TMR0
	
do_TMR1
	BANK1
	bcf		PIE1,TMR1IE				; turn TMR1 int off
	BANK0
	bcf		PIR1,TMR1IF				; clear int flag
	bsf		INTCON,T0IE				; turn TMR0 int on
	movf	cur_servo,W				; get current servo value
	subwf	V255,W					; subtract from 255
	movwf	TMR0					; and start the timer
	bcf		INTCON,T0IF
	SERVO_SET						; set the servo bit
	goto	done					; wait for TMR1 interrupt
	
do_TMR0
	SERVO_CLR						; turn off servo bit
	bcf		INTCON,T0IE				; turn TMR0 int off
	bcf		INTCON,T0IF				; and clear its flag
	movlw	T1MSB					; load TMR1 values
	movwf	TMR1H
	movlw	T1LSB
	movwf	TMR1L
	bcf		PIR1,TMR1IF				; clear TMR1 int flag
	BANK1
	bsf		PIE1,TMR1IE				; enable TMR1 int
	BANK0
	
done
	swapf 	STATUS_TMP,W			; swap STATUS_TEMP register into
	movwf 	STATUS 					; move W into STATUS register
	swapf 	W_TMP,F 				; swap W_TEMP
	swapf 	W_TMP,W 				; swap W_TEMP into W
	retfie							; return from interrupt
	
;==========================================================================
; Setup Code     
;
;  Configure the Outputs, ADC and PIC hardware. The <== lines MUST be done
; when we are using the internal RC oscillator, this sets up the factory
; calibration of the oscillator.  When we set up the ADC module we need
; to make sure that the system has the proper time to correctly read an
; analog value.  In this case we are dividing the 4MHz clock by 8 to get
; a Tad of 1.6us or greater.  This gives us 1/(4MHz/8) = 2us.  Carefully
; read Section 7 of the 12F675 data sheet to understand this.
;
; We're going to user timers and interrupts now.  Timer 0 will be used to 
; set the servo pulse width.  We'll divide the system clock by 8 so that 
; we get about 2.024ms with a full count.  We'll fudge our ADC readings 
; and add 120 to them so that a 0 ADC gives about 1ms pulse and a 255 ADC
; gives about 2ms.  This actually worked pretty well.  Next we'll set Timer 1
; to give us a 19ms period gap between servo pulses.  Yeah, this means that
; our period varies with the pulse width but that doesn't bother servos a
; bit.  Notice that some settings are done in RAM bank 0 and some in RAM
; bank 1.  That is the joy of working with the mid-range PICs, a segmented
; memory structure.
;==========================================================================
init    
    BANK1
	call	0x3FF					; <== Get the cal value
	movwf   OSCCAL          		; <== Store the factory osc. calibration value
	movlw   SCONFIG					; Set how we want to configure I/O pins
	movwf	TRISIO					; Configure input and output pins
	movlw   OPTS		     		; Set OPTION bits as we have defined them
	movwf	OPTION_REG				; Implement OPTION register bits
	movlw	ANALOG
	movwf	ANSEL					; set ADC clock and inputs
	BANK0
	movlw	ANCONF
	movwf	ADCON0					; configure ADC and start converter
	clrf	GPIO					; clear GPIO
	movlw	CMPOFF
	movwf	CMCON					; turn comparitor off
	movlw	T1SETTING				; get TMR1 setup
	movwf	T1CON
	bcf		PIR1,TMR1IF				; clear TMR1 int flag, just in case
	movlw	D'150'					; set some arbitrary servo default
	movwf	servo
	movwf	cur_servo
	movlw	D'105'					; servo center position default
	movwf	TMR0
	clrf	pattern					; clear output bit pattern
	movlw	D'255'
	movwf	V255					; this will be useful later
	bsf		INTCON,T0IE				; enable the TMR0 overflow IRQ
	bsf		INTCON,PEIE				; turn on peripheral ints (subtle bug hint)
	bsf		INTCON,GIE				; turn on the interrupt system
	
;==========================================================================
; Main code loop 
;
;  First take and ADC reading, we're only interested in the upper 8 bits,
; we set the "left justify" bit in ADCON0 so that ADRESH (upper 8 bits of 
; the 10 bit result) gets the analog value divided by 4.  We have to wait 
; for the GO_DONE bit in ADCON0 to clear to show that we have a new value.
; Take this reading and divide that by 2, then add 100 to it to give us
; approximately a count from 100 to 256.  Note the function call set_servo.
; This call only changes the servo settings if they really change, I like
; to avoid unnecessary banging of variables, here is isn't really needed,
; but you can put a filter in this call that will smooth out jitters that
; you might be getting due to noise on your analog input.
;
; As you can see, this main code is pretty thin, all of the heavy lifting
; is done in the "background" by the interrupt system.
;==========================================================================

loop
	ADC_GO							; fire off an ADC sample
	
wait								; stay in this loop until ADCON0, GO_DONE bit clears
	btfsc	ADCON0,GO_DONE
	goto	wait					; not yet
	
	movf	ADRESH,W				; Get the upper 8 bits of the conversion (divide by 4)
	movwf	light					; and save it to the light storage 
	bcf		STATUS,C
	movlw	0xFE					; clear lower nibble
	andwf	light,f					; so that the rrf won't put '1' into the carry bit
	rrf		light,f					; divide the value by 2 
	movlw	D'120'					; base value about 1ms (magic number alert!)
	movwf	servo					; save it
	movf	light,W					; get light value
	addwf	servo,f					; value now is 100 to 228, or 1ms to 2.28ms
	call	set_servo				; set next servo value
	
	goto	loop					; and do it again!

set_servo							; sets up the next servo setting to use
	movf	servo,W					; get the new servo setting
	subwf	cur_servo,W				; compare new and current ones
	btfsc	STATUS,Z				; the same?
	goto	ss_done					; yes, so do nothing
	movf	servo,W					; different, get new value
	movwf	cur_servo				; so we'll use it now
ss_done
	retlw	0						; return from function
		
		END            				; directive 'end of program'

